#!/usr/bin/env python
# -*- coding: utf-8 -*-

#---------------------------------------------------------------------------
#
# レトロフリークカートリッジアダプタ操作プログラム
# (テスト版)
# 
# http://sankichi.backdrop.jp/
#---------------------------------------------------------------------------

# Windows10で実行する際はコマンドプロンプト上で
# > chcp 65001
# を実行してから起動すること。
# また、pyserialライブラリのインストールも必要。

import sys
import glob
import serial
import datetime
import struct
import time

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

#---------------------------------------------------------------------------
# 定数など
#---------------------------------------------------------------------------

# ヘルプメッセージ
helpMsg =   'usage: %s ComPortName                          \n'\
            '       ComPortName : シリアルポート名(例: COM1)'

# トップメニュー
menuStr_Top = \
            '\n'\
            '==========================================================\n'\
            '                     <<トップメニュー>>                   \n'\
            '==========================================================\n'\
            '接続中カセット種別: %s                                    \n'\
            '----------------------------------------------------------\n'\
            '1: 先頭512バイトをダンプ表示                              \n'\
            '2: 指定アドレスから512バイトダンプ表示                    \n'\
            '3: 指定アドレスから指定サイズダンプ表示                   \n'\
            '4: 16MbitメガドラROM読み出し＆ROMファイル作成             \n'\
            '5: 指定サイズROM読み出し＆ROMファイル作成                 \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '\

#---------------------------------------------------------------------------
# 各種処理
#---------------------------------------------------------------------------

# シリアルポート列挙ルーチン
# 以下のサイトを参考にした
# https://codeday.me/jp/qa/20181225/90655.html
def serial_ports():
    """ Lists serial port names

        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result

#---------------------------------------------------------------------------
# byte配列をダンプする
# 参考にしたURL:https://qiita.com/taka_baya/items/c22bd5f5e7cb3de90988
#---------------------------------------------------------------------------
def dump(bytes, addr = 0):
    memory_address = 0 + addr
    ascii_string = ""
    print("Offset   00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F ASCII")
    print("-------- ----------------------------------------------- ----------------")
    
    ## 読み込んだバイナリファイルを表示し終わるまでループします
    for byte in bytes:
        ascii_string = ascii_string + chr(validate_byte_as_printable(byte))
        if memory_address%16 == 0:
            print(format(memory_address, '08X'),end = '')
            print(" " + hex(byte)[2:].zfill(2), end='')
        elif memory_address%16 == 15:
            print(" " + hex(byte)[2:].zfill(2),end='')
            print(" " + ascii_string)
            ascii_string = ""
        else:
            print(" " + hex(byte)[2:].zfill(2), end='')
        memory_address = memory_address + 1

def validate_byte_as_printable(byte):
    ## ascii 文字列化をチェック. asciiでなければ '.' を返す ##
    if is_character_printable(byte):
        return byte
    else:
        return 46

def is_character_printable(s):
    ## asciiの英数字・記号の文字列かどうかを判別し、真偽値を返します。
    #if s < 126 and s >= 33:
    if s < 126 and s >= 0x20:
        return True 


#---------------------------------------------------------------------------
# 指定アドレスから指定サイズデータ読み取り
#       戻り値：2つの値のタプル
#               retCode: 
#                    0:成功
#                   -1:読み取り要求コマンドの応答タイムアウト
#                   -2:読み取り要求コマンドに肯定応答でない応答
#                   -3:読み取りデータ本体受信タイムアウト
#                   -4:ラストの状態要求コマンド応答が不正
#               retAry:  成功の時、読み取ったデータ
#
#       メモ：
#---------------------------------------------------------------------------
def readSize5(s, addrInt, sizeInt):
    retCode = 0
    retAry  = []
    
    addr = addrInt.to_bytes(4, 'little')
    size = sizeInt.to_bytes(4, 'little')

    # 入出力バッファクリア
    if s.in_waiting > 0:
        putLog('未処理の受信データあり: %d' % s.in_waiting)
        readStr = s.read(s.in_waiting)
        readBytes = bytes(readStr)
        putLog('未処理データ受信。')
        debugDump(readBytes)

    else:
        putLog('未受信データ無し。')

    if s.out_waiting > 0:
        putLog('送信バッファにデータあり: %d' % s.out_waiting)
        s.reset_output_buffer()
    else:
        putLog('未送信データ無し。')


    # 1. 読み取り要求コマンドに肯定応答が来るまで待つ処理
    waitCount = 0
    while True:
        # 1-1. 読み取り要求コマンド送信
        putLog('読み取り要求コマンド送信。')
        s.write(bytes([
            0x0D, 0x00, 0x00, 0x00,             # 読み込み
            0x08, 0x00, 0x00, 0x00,             # 不明 8
            size[0], size[1], size[2], size[3], # サイズ？
            addr[0], addr[1], addr[2], addr[3], # 開始アドレス
            size[0], size[1], size[2], size[3], # サイズ？
            ]))

        # 1-2. 読み取り要求コマンド応答待ち
        respWaitCount = 0
        while True:
            if s.in_waiting > 0:
                putLog('受信データあり: %d' % s.in_waiting)
                break
            else:
                respWaitCount = respWaitCount + 1
                putLog('0.01秒待機。')
                time.sleep(0.01)
                if(respWaitCount > 40):
                    print('ERR: 読み取り要求コマンド応答タイムアウト')
                    return -1, retAry

        # 1-3. 読み取り要求コマンド応答の受信と判定
#        readStr = s.read(s.in_waiting)
        readStr = s.read(8)
        readBytes = bytes(readStr)
        putLog('読み取り要求コマンド応答受信。')
        debugDump(readBytes)

        if readBytes == bytes([0,0,0,0,size[0],size[1],size[2],size[3]]):
            putLog('肯定応答を受信。')
            break
        else:
            putLog('肯定応答でない応答を受信。')
            print('ERR: 読み取り要求コマンドに肯定応答でない応答')
            return -2, retAry

    # 2. 状態要求コマンド送信
    # 実機と異なり、ここで、いつまで待ってもデータ本体が来ない。
    # 催促のため(！？)状態要求コマンドを送信する

    # 状態要求コマンド送信
    putLog('読み取りデータ未着のため状態要求コマンド送信。')
    s.write(bytes([
        0x06, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,]))
    # 応答待ち
    waitCount = 0
    while True:
        if s.in_waiting > 0:
            putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            waitCount = waitCount + 1
            if(waitCount > 40):
                retCode = -3
                print('ERR: 読み取りデータ本体受信タイムアウト')
                return retCode, retAry
            putLog('0.01秒待機。')
            time.sleep(0.01)

    # 3. 読み取りデータ本体受信
    #   ただし、前項で状態要求を出していた場合、
    #   読み取りデータに続いて、状態応答が来るので
    #   忘れずに受信する。忘れるとシーケンスが破綻する
    readStr = s.read(sizeInt)
    readBytes = bytes(readStr)
    putLog('読み取りデータ本体受信。')
    debugDump(readBytes, addrInt)
    retAry.extend(readBytes)

    # 4. 状態要求コマンド応答受信
    readStr = s.read(12)
    readBytes = bytes(readStr)
    putLog('状態要求コマンド応答受信。')
    debugDump(readBytes)

    if readBytes == bytes([0xff,0xff,0xff,0xff,0,0,0,0]):
        putLog('ERR: 状態要求コマンド応答が不正。')
        return -4, retAry

    return retCode, retAry


def readSize_dummy(s, addrInt, sizeInt):
    retCode = 0
    retAry  = []
    
    addr = addrInt.to_bytes(4, 'little')
    size = sizeInt.to_bytes(4, 'little')
    
    while True:
        readBytes = [0] * sizeInt
        retAry.extend(readBytes)
        break

    return retCode, retAry

def putLog(msg):
    if verbose == 1:
        print('%s ' % datetime.datetime.now(), end='')
        print(msg)

def debugDump(bytes, addr = 0):
    if verbose == 1:
        dump(bytes, addr)
        print('')

#---------------------------------------------------------------------------
#
# mainルーチン
#
#---------------------------------------------------------------------------
if __name__ == '__main__':
    # 引数の解析
    args = sys.argv
    
    if len(args) < (1 + 1):
        print(helpMsg % (args[0]))
        print('')
        print('使用可能なシリアルポート名一覧:')
        for portName in serial_ports():
            print(portName)
        sys.exit(0)
    
    comPortName = args[1]
    
    try:
        # シリアルポートオープン
        s = serial.Serial(comPortName, 115200, timeout=0.25)

        # ループ
        while True:
            # 状態要求コード(?)送信
            s.write(bytes([
                0x06, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00,
                0x04, 0x00, 0x00, 0x00,]))

            # 応答受信
            readStr = s.read(1024)

            # 応答解析
            readBytes = bytes(readStr)
            #print("%s " % datetime.datetime.now(), end='')
            
            stateStr = ''
            connected = 0
            if len(readBytes) == 0:
                stateStr = 'ERR: カートリッジアダプタが応答しません'
                connected = 0
            elif len(readBytes) < 12:
                stateStr = 'ERR: カートリッジアダプタの応答が不正です。'
                connected = 0
            else:
                if   readBytes[8] == 0xff:
                    stateStr = '未接続'
                    connected = 0

                elif readBytes[8] == 0x01:
                    stateStr = 'スーパーファミコン'
                    connected = 1

                elif readBytes[8] == 0x02:
                    stateStr = 'メガドライブ'
                    connected = 1

                elif readBytes[8] == 0x03:
                    stateStr = 'ファミコン'
                    connected = 1

                elif readBytes[8] == 0x05:
                    stateStr = 'ゲームボーイ'
                    connected = 1

                elif readBytes[8] == 0x08:
                    stateStr = 'PCエンジンHuカード'
                    connected = 1

            # メニュー表示
            print(menuStr_Top % stateStr, end='')
            
            # 入力待ち＆判定
            inStr = input()
            if len(inStr) == 0:
                continue

            elif inStr == '1':      # 先頭512バイトダンプ
                if connected != 1:
                    print('カセットが未接続のためダンプできません。')
                else:
                    print('先頭ダンプ')
                    (retCode, retAry) = readSize5(s, 0, 512)
                    print('retCode: %d' % retCode)
                    dump(retAry)
                    print('')
                continue

            elif inStr == '2':      # 指定アドレス512バイトダンプ
                if connected != 1:
                    print('カセットが未接続のためダンプできません。')
                else:
                    print('指定アドレスダンプ')

                    # アドレスを入力させる
                    while True:
                        print('開始アドレス(16進)を入力してください。: ', end='')
                        stAddrStr = input()
                        if len(stAddrStr) == 0:
                            continue
                        else:
                            if len(stAddrStr) > 2 and stAddrStr.startswith('0x'):
                                stAddr = int(stAddrStr, 0)
                            else:
                                stAddr = int(stAddrStr, 16)
                        break

                    (retCode, retAry) = readSize5(s, stAddr, 512)
                    dump(retAry, stAddr)
                    print('')
                continue

            elif inStr == '3':      # 指定アドレス指定サイズダンプ
                if connected != 1:
                    print('カセットが未接続のためダンプできません。')
                else:
                    putLog('指定アドレス指定サイズダンプ')

                    # アドレスを入力させる
                    while True:
                        print('開始アドレス(16進)を入力してください。: ', end='')
                        stAddrStr = input()
                        if len(stAddrStr) == 0:
                            continue
                        else:
                            if len(stAddrStr) > 2 and stAddrStr.startswith('0x'):
                                stAddr = int(stAddrStr, 0)
                            else:
                                stAddr = int(stAddrStr, 16)
                        break

                    # サイズを入力させる
                    while True:
                        print('サイズ(16進)を入力してください。: ', end='')
                        stSizeStr = input()
                        if len(stSizeStr) == 0:
                            continue
                        else:
                            if len(stSizeStr) > 2 and stSizeStr.startswith('0x'):
                                stSize = int(stSizeStr, 0)
                            else:
                                stSize = int(stSizeStr, 16)
                        break

                    (retCode, retAry) = readSize5(s, stAddr, stSize)
                    dump(retAry, stAddr)
                    print('')
                continue

            elif inStr == '4':      # 16Mbitメガドラカセットファイル化
                if connected != 1:
                    print('カセットが未接続のためダンプできません。')
                else:
                    putLog('16Mbitメガドラカセットファイル化')

                    # ファイル名入力
                    while True:
                        print('作成ファイル名を入力してください。: ', end='')
                        fileNameStr = input()
                        if len(fileNameStr) == 0:
                            continue
                        break

                    fileAry = []
                    addr = 0
                    dumpSize = 1024
                    #dumpSize = 512
                    while addr < 1024*1024*2:
                        # ファイル出力用配列にバンバン追加
                        print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                        (retCode, retAry) = readSize5(s, addr, dumpSize)
                        if retCode == 0:
                            fileAry.extend(retAry)
                            addr = addr + dumpSize
                        else:
                            print('リトライ実施(retCode: %d)' % retCode)
                            #s.rtscts = 1
                            continue
                    print('完了。')
                    
                    # 配列をファイル出力
                    with open(fileNameStr, "wb") as fOut:
                        for x in fileAry:
                            fOut.write(struct.pack("B", x))
                    print('ファイル: %sを出力しました。' % fileNameStr)
                    print('')
                continue

            elif inStr == '5':      # 指定サイズメガドラカセットファイル化
                if connected != 1:
                    print('カセットが未接続のためダンプできません。')
                else:
                    putLog('指定サイズメガドラカセットファイル化')

                    # ファイル名入力
                    while True:
                        print('作成ファイル名を入力してください。: ', end='')
                        fileNameStr = input()
                        if len(fileNameStr) == 0:
                            continue
                        break

                    # サイズ入力
                    while True:
                        print('サイズ(16進)を入力してください。: ', end='')
                        endAddrStr = input()
                        if len(endAddrStr) == 0:
                            continue
                        else:
                            if len(endAddrStr) > 2 and endAddrStr.startswith('0x'):
                                endAddr = int(endAddrStr, 0)
                            else:
                                endAddr = int(endAddrStr, 16)
                        break

                    fileAry = []
                    addr = 0
                    dumpSize = 1024
                    #dumpSize = 512
                    while addr < endAddr:
                        # ファイル出力用配列にバンバン追加
                        print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                        (retCode, retAry) = readSize5(s, addr, dumpSize)
                        if retCode == 0:
                            fileAry.extend(retAry)
                            addr = addr + dumpSize
                        else:
                            print('リトライ実施(retCode: %d)' % retCode)
                            #s.rtscts = 1
                            continue
                    print('完了。')
                    
                    # 配列をファイル出力
                    with open(fileNameStr, "wb") as fOut:
                        for x in fileAry:
                            fOut.write(struct.pack("B", x))
                    print('ファイル: %sを出力しました。' % fileNameStr)
                    print('')
                continue

            elif inStr == 'q':      # 終了
                print('終了します。')
                break

            else:                   # 不明コマンド
                print('不明なコマンドです。')
                continue
        
        # End of while
        
        # シリアルポートクローズ
        s.close()
    except (OSError, serial.SerialException):
        print('例外発生')
        pass
    
    
    #print(serial_ports())
